///////////////////////////////////////////////////////////
// Complex FIR Filter example for Xilinx MicroBlaze FSL
// 
// Created by Bruce Karsten of Xilinx (bruce.karsten@xilinx.com)
//

#include <stdio.h>
#include "co.h"
#include "cosim_log.h"
#include "Filter.h"

extern void call_accelerator (co_stream output_stream, co_stream input_stream);

void complex_fir (co_stream filter_in, co_stream filter_out) {
	int32	coef_mem[IF_FILT_LEN];
	int32	filt_hist[IF_FILT_LEN];
	int32	inSample;
	int32	outFilter;
	int		i;
	int		write_idx;
	int		read_idx;
	int		t_read_idx, t_read_idx2;
	int32	t_coef, t_hist;
	int16	coef_real, coef_imag;
	int16	hist_real, hist_imag;
	int32	p1_real, p2_real;
	int32	p1_imag, p2_imag;
	int32	res_real, res_imag;
	int32	accum_real, accum_imag;
	IF_SIM (int samplesread = 0;)
	IF_SIM (int sampleswritten = 0;)
	
	IF_SIM ( cosim_logwindow log; )
	IF_SIM ( log = cosim_logwindow_create ("complex_fir"); )

	co_array_config(coef_mem,co_kind,"async");
	co_array_config(filt_hist,co_kind,"async");
	
	do { // Hardware processes run forever
	
		co_stream_open (filter_in,  O_RDONLY, INT_TYPE(32));
		co_stream_open (filter_out, O_WRONLY, INT_TYPE(32));
		
		// Init history pointers
		write_idx = IF_FILT_LEN - 1;
		read_idx = 0;

		i = 0;
		do {
#pragma CO PIPELINE	
			// Fill the Coefficient Memory
			co_stream_read (filter_in, &inSample, sizeof(int32));
			coef_mem[i] = inSample;
			// Clear the filter history
			filt_hist[i++] = 0;
		} while(i < IF_FILT_LEN);

		// Begin to process incoming data
		while (co_stream_read (filter_in, &inSample, sizeof(int32)) == co_err_none) {
			IF_SIM (samplesread++;)

			filt_hist[write_idx] = inSample;
			if (write_idx == (IF_FILT_LEN - 1))
					write_idx = 0;
				else
					write_idx++;

			t_read_idx = read_idx++;
			if (read_idx >= IF_FILT_LEN )
					read_idx = 0;

			//
			// Create Filtered value
			//
			
			// Clear the accums
			accum_real = 0;
			accum_imag = 0;

			for (i = 0;i < IF_FILT_LEN;i++){ 
#pragma CO PIPELINE	
#pragma CO set stageDelay 32
				// Get the pieces
				t_coef = coef_mem[i];
				t_hist = filt_hist[t_read_idx++];

				if (t_read_idx >= IF_FILT_LEN )
					t_read_idx = 0;

				coef_real = (t_coef >> 16) & 0xFFFF;
				coef_imag = (t_coef & 0xFFFF);
				hist_real = (t_hist >> 16) & 0xFFFF;
				hist_imag = (t_hist & 0xFFFF);

				// Create the products
				p1_real = coef_real * hist_real;
				p1_imag = coef_real * hist_imag;
				p2_real = coef_imag * hist_imag;
				p2_imag = coef_imag * hist_real;

				// Combine products & reduce bits
				res_real = p1_real - p2_real;
				res_imag = p1_imag + p2_imag;
				res_real >>= 12;
				res_imag >>= 12;
				
				// Accumulate products
				accum_real += res_real;
				accum_imag += res_imag;
			}

			// Return result to 16 bits (assuming > 256 length filter)
			accum_real >>= 13;
			accum_imag >>= 13;
			
			// Transfer result to output word
			outFilter = ((accum_real & 0xFFFF) << 16) + (accum_imag & 0xFFFF);
				
			// Output Filtered value
			co_stream_write (filter_out, &outFilter, sizeof(int32));
			IF_SIM(sampleswritten++;)			
		}
		IF_SIM (cosim_logwindow_fwrite (log, 
			"Closing Complex FIR, samples read : %d, samples written : %d\n",
			samplesread, sampleswritten);)
			
		co_stream_close (filter_in);
		co_stream_close (filter_out);
		
		IF_SIM(break;)
	} while (1);
}

void config_filt (void *arg) {
	int i;
	
	co_stream  to_filt, from_filt;
	co_process cpu_proc, filter_proc;
	
	to_filt   = co_stream_create ("to_filt",   INT_TYPE(32), 4); 
	from_filt = co_stream_create ("from_filt", INT_TYPE(32), 4);
	
	cpu_proc    = co_process_create ("cpu_proc",    (co_function) call_accelerator, 2, to_filt, from_filt);
	filter_proc = co_process_create ("filter_proc", (co_function) complex_fir,      2, to_filt, from_filt);
	
	co_process_config (filter_proc, co_loc, "PE0");
}

co_architecture co_initialize () {
	return (co_architecture_create ("filt", "xilinx_apu", config_filt, NULL));
}
